/*************************************************************************
	PlayStation Low Level Camera routines
 
	by Christopher Deel 
			
		Copyright (C) 1995,96 Gametek Inc.
		All rights Reserved
 
*************************************************************************/

#include <sys/types.h>

#include <libetc.h>			   /* For Controller routines */
#include <libgte.h>			   /* For LIBGS */
#include <libgpu.h>			   /* For LIBGS */
#include <libgs.h>			   /* Graphics Library */
#include <libsn.h>

#include "prjtypes.h"
#include "prjhdrs.h"    /* Project header files */
#include "debug.h"
#include "Camera.h"

#include "fbequ.h"
#include "fbram.h"

Camera_t gCamera, gCameraSaved;


void DumpCamera(void);


/*-------------------------------------------------------------
  CalcCameraMoveValue ~ Calc movement of camera to target value
  		Speed must be positive
-------------------------------------------------------------*/
int CalcCameraMoveValue(int pos, int target, int speed)
{
	if (pos == target)
		return pos;
		
	if (pos < target)
	{
		if ( (target - pos) < speed)
			pos = target;
		else
			pos = pos + speed;
	}
	else
	{
		if ((pos - target) < speed)
			pos = target;
		else
			pos = pos - speed;
	}
	return pos;
}

/*-------------------------------------------------------------
  SetCameraOffset ~ Set Camera offset from Viewpoint
-------------------------------------------------------------*/
void SetCameraOffset(int x, int y, int z)
{
	
	if (!x && !y && !z)
	{ 
	 	 Cerror();
	 	 DumpCamera();    
		return; 	 // error
	}
	if (gCamera.flags & CAMERA_FLIPZ_FLAG)
	{
		z = -z;
		x = -x;   // test
	}
		
	if (gCamera.flags & CAMERA_OFFSET_TARGET_FLAG)
	{
		gCamera.offsetTarget.vx = x;
		gCamera.offsetTarget.vy = y;
		gCamera.offsetTarget.vz = z;
	}
	else
	{
		gCamera.offset.vx = x;
		gCamera.offset.vy = y;
		gCamera.offset.vz = z;
	}
//	 if (AbsInt(z) > 6000) 	DumpCamera();  // debug
}

/*-------------------------------------------------------------
  SetCameraOffsetNoFlip ~ Set Camera offset from Viewpoint
    No Flip calc
-------------------------------------------------------------*/
void SetCameraOffsetNoFlip(int x, int y, int z)
{
	if (gCamera.flags & CAMERA_OFFSET_TARGET_FLAG)
	{
		gCamera.offsetTarget.vx = x;
		gCamera.offsetTarget.vy = y;
		gCamera.offsetTarget.vz = z;
	}
	else
	{
		gCamera.offset.vx = x;
		gCamera.offset.vy = y;
		gCamera.offset.vz = z;
	}
//	 DumpCamera();    // debug
}

/*-------------------------------------------------------------
  SetCameraOffsetVector ~ 
-------------------------------------------------------------*/
void SetCameraOffsetVector(SVECTOR *vec)
{
	SetCameraOffset(vec->vx, vec->vy, vec->vz);
}

/*-------------------------------------------------------------
  MoveCameraOffset ~ Move Camera offset
-------------------------------------------------------------*/
void MoveCameraOffset(int x, int y, int z)
{
	gCamera.offset.vx += x;
	gCamera.offset.vy += y;
	gCamera.offset.vz += z;
	limitRange(gCamera.offset.vx, gCamera.minOffset.vx, gCamera.maxOffset.vx);
	limitRange(gCamera.offset.vy, gCamera.minOffset.vy, gCamera.maxOffset.vy);
	limitRange(gCamera.offset.vz, gCamera.minOffset.vz, gCamera.maxOffset.vz);
	copyVector(&gCamera.offsetTarget, &gCamera.offset);
}

/*-------------------------------------------------------------
  SetCameraMovement ~ Set constant movement of camera
-------------------------------------------------------------*/
void SetCameraMovement(int x, int y, int z )
{
	gCamera.offsetMovement.vx = x;
	gCamera.offsetMovement.vy = y;
	gCamera.offsetMovement.vz = z;
}

/*-------------------------------------------------------------
  SetCameraMax ~ Set Max offset for camera
-------------------------------------------------------------*/
void SetCameraMax(int x, int y, int z)
{
	gCamera.maxOffset.vx = x;
	gCamera.maxOffset.vy = y;
	gCamera.maxOffset.vz = z;
}

/*-------------------------------------------------------------
  SetCameraMin ~ Set minimum offset for camera
-------------------------------------------------------------*/
void SetCameraMin(int x, int y, int z)
{
	gCamera.minOffset.vx = x;
	gCamera.minOffset.vy = y;
	gCamera.minOffset.vz = z;
}

/*-------------------------------------------------------------
  SetCameraReference ~ Set refernce point of camera
-------------------------------------------------------------*/
void SetCameraReference(int x, int y, int z)
{
	register int rx, ry, rz;
	register int sx, sy, sz;
	
	rx = min(x, gCamera.refMax.vx);
 	ry = min(y, gCamera.refMax.vy);
 	rz = min(z, gCamera.refMax.vz);
	gCamera.ref.vx = max(rx, gCamera.refMin.vx);
 	gCamera.ref.vy = max(ry, gCamera.refMin.vy);
 	gCamera.ref.vz = max(rz, gCamera.refMin.vz);
	
	sx = AbsInt(rx - view.vrx) / 100;
	sy = AbsInt(ry - view.vry) / 100;
	sz = AbsInt(rz - view.vrz) / 100;
	limitRange(sx, 4, 60);
	limitRange(sy, 4, 60);
	limitRange(sz, 4, 60);
	gCamera.refMinSpeed.vx = sx;
	gCamera.refMinSpeed.vy = sy;
	gCamera.refMinSpeed.vz = sz;
}

/*-------------------------------------------------------------
  SetCameraReferenceVector ~ 
-------------------------------------------------------------*/
void SetCameraReferenceVector(SVECTOR *vec)
{
	SetCameraReference(vec->vx, vec->vy, vec->vz);
}  /* End SetCameraReferenceVector() */

/*-------------------------------------------------------------
  SetCameraRefMax ~ 
-------------------------------------------------------------*/
void SetCameraRefMax(int x, int y, int z)
{
	gCamera.refMax.vx = x;
	gCamera.refMax.vy = y;
	gCamera.refMax.vz = z;
}

/*-------------------------------------------------------------
  SetCameraRefMin ~ 
-------------------------------------------------------------*/
void SetCameraRefMin(int x, int y, int z)
{
	gCamera.refMin.vx = x;
	gCamera.refMin.vy = y;
	gCamera.refMin.vz = z;
}

/*-------------------------------------------------------------
  GetCameraVericalCameraAngle ~ 
-------------------------------------------------------------*/
long GetVerticalCameraAngle(void)
{
	int a,b, radius;
	long angle;
	
	a = abs(gCamera.offsetTarget.vx);
	b = abs(gCamera.offsetTarget.vz);
	radius = SquareRoot0((a*a) + (b*b));
	angle = ratan2(abs(gCamera.offsetTarget.vy), radius);
	return angle;
} 

/*-------------------------------------------------------------
  GetCameraAngleXZ ~ 
-------------------------------------------------------------*/
long GetCameraAngleXZ(void)
{
	register int angle;
	
	angle = ratan2(gCamera.offsetTarget.vx, gCamera.offsetTarget.vz);
	return angle;
}  

/*-------------------------------------------------------------
  SetCameraAngle ~ Set Camera XZ angle
-------------------------------------------------------------*/
void SetCameraAngle(int angle)
{
	register int x,z;
	
	gCamera.angle = angle;
	x = (csin(angle) * gCamera.radius) /4096;
	z = (ccos(angle) * gCamera.radius) /4096;
	SetCameraOffset(x, gCamera.offsetTarget.vy, z);
}

/*-------------------------------------------------------------
  SetCameraRadius ~ 
-------------------------------------------------------------*/
void SetCameraRadius(int radius)
{
	gCamera.radius = radius;
}

/*-------------------------------------------------------------
  CalcCameraAngle ~ Calculate current XZ camera angle
  			( in PlayStation format )
-------------------------------------------------------------*/
int CalcCameraAngle(void)
{
	register int angle;
	
	angle = ratan2(gCamera.offsetTarget.vx, gCamera.offsetTarget.vz);
	if (angle < 0) 
		angle += 4096;
	gCamera.angle = angle;	
	return angle;
}

/*-------------------------------------------------------------
  MoveCameraAngle ~ 
-------------------------------------------------------------*/
void MoveCameraAngle(int move)
{
	int angle;
	
//	angle = GetCameraAngle();
	angle = gCamera.angle;
	SetCameraAngle(angle + move);
//	 DumpCamera();
//	 DebugPrintf(0, "Angle %d (%d,%d)\n", angle, 
//			 gCamera.offsetTarget.vx, gCamera.offsetTarget.vz);
//	 VSync(0);		 
}

/*-------------------------------------------------------------
  CalcCameraRadius ~ 
-------------------------------------------------------------*/
int CalcCameraRadius(void)
{
	register int a,b,c;
	
	a = abs(gCamera.offsetTarget.vx);
	b = abs(gCamera.offsetTarget.vz);
	c = SquareRoot0((a*a) + (b*b));
	gCamera.radius = c;
	return c;
}

/*-------------------------------------------------------------
-------------------------------------------------------------*/
static int CalcNewReferencePos(int pos, int target, 
								int minspeed, int maxspeed)
{
	int move, ax, diff ;
	
	diff = target - pos;
	move = diff / gCamera.refDiv;
	
	if (move == 0 || (AbsInt(diff) < minspeed) )
	{
		pos = target;
	}
	else
	{
		ax = AbsInt(move);
				
		if (ax < minspeed)
		{
			if (move > 0)
				move = minspeed; 
			else
				move = -minspeed;
		}
		
		if (ax > maxspeed)
			pos = CalcCameraMoveValue(pos, target, maxspeed);
		else
			pos += move;
	}			
	return pos;
}  /* End CalcReferenceMoveValue() */


/*-------------------------------------------------------------
  DoCamera ~ 
-------------------------------------------------------------*/
void DoCamera(void)
{
	register Camera_t *cameraPtr;
	register int x,y,z;   // for speed
	
	cameraPtr = &gCamera;
	if (cameraPtr->flags & CAMERA_INSTANT_FLAG)
	{
		cameraPtr->zoom = cameraPtr->zoomTarget;
		cameraPtr->zoomPrev = cameraPtr->zoom;
		GsSetProjection(cameraPtr->zoom);
		copyVector(&cameraPtr->offset, &cameraPtr->offsetTarget);
		view.vrx = cameraPtr->ref.vx;
		view.vry = cameraPtr->ref.vy;
		view.vrz = cameraPtr->ref.vz;
		ResetBit(cameraPtr->flags, CAMERA_INSTANT_FLAG);
	}
	if (cameraPtr->zoom != cameraPtr->zoomTarget)
	{
		cameraPtr->zoom = CalcCameraMoveValue(cameraPtr->zoom, cameraPtr->zoomTarget,
									cameraPtr->zoomSpeed);
	}
	if (cameraPtr->zoom != cameraPtr->zoomPrev)
	{
		GsSetProjection(cameraPtr->zoom);
	}
	
	if (cameraPtr->refDiv < 1)
	{
		view.vrx = cameraPtr->ref.vx;
		view.vry = cameraPtr->ref.vy;
		view.vrz = cameraPtr->ref.vz;
	}
	else
	{
		register int speed, ax,ay,az;
		
		speed = cameraPtr->refSpeed;
		view.vrx = CalcNewReferencePos(view.vrx, cameraPtr->ref.vx,
						cameraPtr->refMinSpeed.vx, speed);
						
		view.vry = CalcNewReferencePos(view.vry, cameraPtr->ref.vy,
						cameraPtr->refMinSpeed.vy, speed);
						
		view.vrz = CalcNewReferencePos(view.vrz, cameraPtr->ref.vz,
						cameraPtr->refMinSpeed.vz, speed);

#if 0				
		x = (cameraPtr->ref.vx - view.vrx) / cameraPtr->refDiv;
		y = (cameraPtr->ref.vy - view.vry) / cameraPtr->refDiv;
		z = (cameraPtr->ref.vz - view.vrz) / cameraPtr->refDiv;
		ax = AbsInt(x);
		ay = AbsInt(y);
		az = AbsInt(z);
				
		if (ax < cameraPtr->refMinSpeed.vx)
			x = cameraPtr->refMinSpeed.vx * sign(x);
		if (ay < cameraPtr->refMinSpeed.vy)
			y = cameraPtr->refMinSpeed.vy * sign(y);
		if (az < cameraPtr->refMinSpeed.vz)
			z = cameraPtr->refMinSpeed.vz * sign(z);
		
		if (ax > speed)
			view.vrx = CalcCameraMoveValue(view.vrx, cameraPtr->ref.vx, speed);
//		else if (ax < cameraPtr->refMinSpeed.vx)
//			view.vrx = CalcCameraMoveValue(view.vrx, cameraPtr->ref.vx, cameraPtr->refMinSpeed.vx);
		else
			view.vrx += x;
		
		if (ay > speed)
			view.vry = CalcCameraMoveValue(view.vry, cameraPtr->ref.vy, speed);
//		else if (ay < cameraPtr->refMinSpeed.vy)
//			view.vry = CalcCameraMoveValue(view.vry, cameraPtr->ref.vy, cameraPtr->refMinSpeed.vy);
		else
			view.vry += y;
			
		if (az > speed)
			view.vrz = CalcCameraMoveValue(view.vrz, cameraPtr->ref.vz, speed);
//		else if (az < cameraPtr->refMinSpeed.vz)
//			view.vrz = CalcCameraMoveValue(view.vrz, cameraPtr->ref.vz, cameraPtr->refMinSpeed.vz);
		else
			view.vrz += z;
#endif // 0	

	}
	// Offset Calc
	x = cameraPtr->offset.vx;
	y = cameraPtr->offset.vy;
	z = cameraPtr->offset.vz;
	if (cameraPtr->flags & CAMERA_OFFSET_TARGET_FLAG)
	{
		if (x == cameraPtr->offsetTarget.vx 
				&& y == cameraPtr->offsetTarget.vy
				&& z == cameraPtr->offsetTarget.vz)
		{
			SetBit(cameraPtr->flags, CAMERA_AT_TARGET_FLAG);
		}
		else
		{
			ResetBit(cameraPtr->flags, CAMERA_AT_TARGET_FLAG);
			x = CalcCameraMoveValue(x , cameraPtr->offsetTarget.vx, 
									abs(cameraPtr->offsetMovement.vx) );
			y = CalcCameraMoveValue(y , cameraPtr->offsetTarget.vy, 
									abs(cameraPtr->offsetMovement.vy) );
			z = CalcCameraMoveValue(z , cameraPtr->offsetTarget.vz, 
									abs(cameraPtr->offsetMovement.vz) );
		}
	}
	else
	{
	 	x += cameraPtr->offsetMovement.vx;
		y += cameraPtr->offsetMovement.vy;
		z += cameraPtr->offsetMovement.vz;
	}
	limitRange(x, cameraPtr->minOffset.vx, cameraPtr->maxOffset.vx);
	limitRange(y, cameraPtr->minOffset.vy, cameraPtr->maxOffset.vy);
	limitRange(z, cameraPtr->minOffset.vz, cameraPtr->maxOffset.vz);
//	x = min(x, cameraPtr->maxOffset.vx);
//	y = min(y, cameraPtr->maxOffset.vy);
//	z = min(z, cameraPtr->maxOffset.vz);
//	x = max(x, cameraPtr->minOffset.vx);
//	y = max(y, cameraPtr->minOffset.vy);
//	z = max(z, cameraPtr->minOffset.vz);
	cameraPtr->offset.vx = x;
	cameraPtr->offset.vy = y;
	cameraPtr->offset.vz = z;

	view.vpx = view.vrx + x;
	view.vpy = view.vry + y;
	view.vpz = view.vrz + z;
	view.vpy = min(view.vpy, cameraPtr->viewMax.vy);
	view.vpy = max(view.vpy, cameraPtr->viewMin.vy);
	
	// Calc Direction
	x = view.vrz - view.vpz;
	y = view.vrx - view.vpx;
	cameraPtr->dir = FindDirection(x, y);
	cameraPtr->refPrev.vx = view.vrx;
	cameraPtr->refPrev.vy = view.vry;
	cameraPtr->refPrev.vz = view.vrz;
	cameraPtr->zoomPrev = cameraPtr->zoom;
}

/*-------------------------------------------------------------
  InitCamera ~ Init camera values
-------------------------------------------------------------*/
void InitCamera(void)
{
	SetBit(gCamera.flags, CAMERA_OFFSET_TARGET_FLAG);
	SetCameraOffset(0,-750, -2500);
	SetCameraMovement(0,0,0);
	SetCameraMax(11000,11000,11000);
	SetCameraMin(-11000,-11000,-11000);
	SetCameraRefMax(31000, 31000, 31000);
	SetCameraRefMin(-31000, -31000, -31000);
	gCamera.viewMin.vy = -4200;
	gCamera.viewMax.vy = -200;
	gCamera.refDiv = 0;
	gCamera.zoom = 1000;
	gCamera.zoomTarget = gCamera.zoom;
	gCamera.zoomSpeed = 50;
	gCamera.refSpeed = 200;
	setVector(&gCamera.refMinSpeed, 4,4,4);
}

/*-------------------------------------------------------------
  DoCameraJoypad ~ 
-------------------------------------------------------------*/
void DoCameraJoypad(void)
{
	ResetBit(gCamera.flags, CAMERA_OFFSET_TARGET_FLAG);
	SetCameraMax(22000,22000,22000);
	SetCameraMin(-22000,-22000,-22000);
	SetCameraRefMax(31000, 31000, 31000);
	SetCameraRefMin(-31000, -31000, -31000);
	SetCameraMovement(0,0,0);
	gCamera.refDiv = 2;
	GetPadxyz(0, 1, 32, &gCamera.offset);
	GetPadxyz(0, 0, 32, &gCamera.ref);
	SetCameraOffset(gCamera.offset.vx, gCamera.offset.vy, gCamera.offset.vz);
	SetCameraReference(gCamera.ref.vx, gCamera.ref.vy, gCamera.ref.vz);
	DoCamera();
}

/*-------------------------------------------------------------
  SaveCamera ~ Save Camera Variables
-------------------------------------------------------------*/
void SaveCamera(void)
{
	CopyBytes((UBYTE*)&gCamera, (UBYTE*)&gCameraSaved, sizeof(Camera_t));
}

/*-------------------------------------------------------------
  RestoreCamera ~ 
-------------------------------------------------------------*/
void RestoreCamera(void)
{
	CopyBytes((UBYTE*)&gCameraSaved, (UBYTE*)&gCamera, sizeof(Camera_t));
	view.vrx = gCamera.ref.vx;
	view.vry = gCamera.ref.vy;
	view.vrz = gCamera.ref.vz;
	view.vpx = view.vrx + gCamera.offset.vx;
	view.vpy = view.vry + gCamera.offset.vy;
	view.vpz = view.vrz + gCamera.offset.vz;
	gCamera.refPrev.vx = view.vrx;
	gCamera.refPrev.vy = view.vry;
	gCamera.refPrev.vz = view.vrz;
}

/*-------------------------------------------------------------
  MoveCameraInstant(); ~ 
-------------------------------------------------------------*/
void MoveCameraInstant(void)
{
	SetBit(gCamera.flags, CAMERA_INSTANT_FLAG);
}



/*********************** Debug routines ******************/
/*-------------------------------------------------------------
  DumpCamera ~
-------------------------------------------------------------*/
void DumpCamera(void)
{
#if PC_DEBUG
	DebugPrintf(0, "Cam: O(%d,%d,%d) ",
			gCamera.offset.vx, gCamera.offset.vy, gCamera.offset.vz);
	DebugPrintf(0, "T(%d,%d,%d)  ",
			gCamera.offsetTarget.vx, gCamera.offsetTarget.vy,
										gCamera.offsetTarget.vz);
	DebugPrintf(0, "Flags:%X\n", gCamera.flags);
	
	DebugPrintf(0, "   R(%d,%d,%d) ", view.vrx, view.vry, view.vrz);
	DebugPrintf(0, " Rt(%d,%d,%d) ", gCamera.ref.vx, gCamera.ref.vy, gCamera.ref.vz);
	DebugPrintf(0, "Dir:%d", gCamera.dir);
	DebugPrintf(0, "\n");
	
	DebugPrintf(0, "  Max(%d,%d,%d)  ",
			 gCamera.maxOffset.vx,gCamera.maxOffset.vy,gCamera.maxOffset.vz);
	DebugPrintf(0, "  Min(%d,%d,%d)",
			gCamera.minOffset.vx,gCamera.minOffset.vy,gCamera.minOffset.vz);
	DebugPrintf(0, "\n");
	DebugPrintf(0, "OnRef %d\n", CameraOnRefMacro);
#endif	
}
