/*--------------------------------------------------------------------------
	MEMORY.C  Replacements for the standard PSX heap management functions.         **
 	By John Brandwood

 	Edited by Christopher Deel

--------------------------------------------------------------------------*/


// **************************************************************************
// **************************************************************************
// **                                                                      **
// ** PSXMem.c                                                      MODULE **
// **                                                                      **
// ** Replacements for the standard PSX heap management functions.         **
// **                                                                      **
// ** By John Brandwood                                                    **
// **                                                                      **
// **   *** WARNING ***                                                    **
// **   *** WARNING ***                                                    **
// **   *** WARNING ***                                                    **
// **                                                                      **
// **   This code MUST be compiled with the "-O2" or "-O3" option since    **
// **   the PsyQ startup and library code are expecting to call these      **
// **   routines with their arguments passed only in registers.            **
// **                                                                      **
// **   *** WARNING ***                                                    **
// **   *** WARNING ***                                                    **
// **   *** WARNING ***                                                    **
// **                                                                      **
// ** Release 1.1 (16Dec95) - Added extra info to DumpHeap().              **
// ** Release 1.2 (16Dec95) - Added extra DEBUG dependant checks for an    **
// **                         attempt to free memory twice, and also for   **
// **                         corrupted or invalid memory blocks.          **
// ** Release 1.3 (04Mar96) - Removed some of the code that was added in   **
// **                         earlier to prove that these routines work.   **
// **                         Dramatically improved the speed of calloc.   **
// ** Release 1.4 (03May96) - Rewritten to improve detection of heap       **
// **                         corruption and to store the file, line       **
// **                         and timecount when each block is allocated.  **
// **                         Routines now disable interrupts while        **
// **                         executing to protect themselves from a       **
// **                         multi-tasking environment.                   **
// **                         Stack pointer is no longer overridden.       **
// ** Release 1.5 (20May96) - Changed memory size settings to use LIBSN's  **
// **                         current way of doing things. This makes the  **
// **                         code slightly more OS-legal.                 **
// ** Release 1.6 (24May96) - Changed interrupt macros to make them GNU    **
// **                         legal and inform the compiler what is going  **
// **                         on. Not 100% sure it is needed, but it does  **
// **                         eliminate a potential problem.               **
// **                                                                      **
// ** N.B.  To take advantage of these routines' ability to store the file **
// **       and line of each malloc(), you will need to either change all  **
// **       your memory allocations to use mallocfl() and callocfl() or    **
// **       use #define directives like these ...                          **
// **                                                                      **
// **       #define malloc(xx)     mallocfl(xx,__FILE__,__LINE__)          **
// **       #define calloc(xx,yy)  callocfl(xx,yy,__FILE__,__LINE__)       **
// **       #define realloc(xx,yy) reallocfl(xx,yy,__FILE__,__LINE__)      **
// **                                                                      **
// **************************************************************************
// **************************************************************************

#include <stdlib.h>

#include "version.h"	// For PC_DEBUG  (Chris Deel)


#if PC_DEBUG
// Define the global DEBUG flag if it hasn't already been defined.
#ifndef DEBUG
  #define DEBUG  1
#endif
#endif PC_DEBUG

//#define REALLOC_USED 1
//#define CALLOC_USED 1

//
// Memory Size Definitions.
//
// It is no longer necessary to link with 2mbyte.obj or 8mbyte.obj to set
// up your program's memory size.
//
// Instead, you now only need to link in LIBSN.LIB and define the variables
// _ramsize and _stacksize.
//
// The MEM_END definition used in the 1.4 release of this code has been
// kept, but it now defaults to zero. Definitions for _ramsize and
// _stacksize have been added to allow the memory to be set up in the
// officially supported way.
//
// If MEM_END is defined as non-zero, then it will still override the
// memory size parameter that is passed to InitHeap().
//

#if DEBUG
  // STANDARD DEBUGGING MODE.
  //
  // Allow for a 4KB stack (change _stacksize to allow for more).
  //
int _ramsize              = 0x00200000;
int _stacksize            = 0x00001000;
  #define MEM_END             0x00000000
#else
  // FINAL PRODUCTION MODE.
  //
  // Allow for a 4KB stack (change _stacksize to allow for more).
  //
int _ramsize              = 0x00200000;
int _stacksize            = 0x00001000;
  #define MEM_END             0x00000000
#endif

//
// Global Functions
//

extern	int                 TestHeap            (void);
extern	void                DumpHeap            (void);

extern	void                InitHeap            (unsigned long *, unsigned long);
extern	void *              malloc              (size_t);
extern	void *              calloc              (size_t, size_t);
extern	void *              realloc             (void *, size_t);
extern	void *              mallocfl            (size_t, char *, unsigned long);
extern	void *              callocfl            (size_t, size_t, char *, unsigned long);
extern	void *              reallocfl           (void *, size_t, char *, unsigned long);
extern	void                free                (void *);

//
// Static Definitions.
//

// The contents of a MEMCHUNK_T header depend upon whether the chunk is
// free or in use.
//
// next -> if free ... ptr to next free block
//         if used ... ptr to filename where malloc was called
// type -> if free ... (+ve) for future expansion
//         if used ... (-ve) TYPE_USED + timestamp + line number of malloc
// size -> both ...... size of block in MEMCHUNK_T units
// csum -> both ...... checksum of the header

struct MEMCHUNK_S;

typedef union MEMPTR_U
{
	struct MEMCHUNK_S * next;
	char *              file;
	unsigned long       data;
} MEMPTR_T;

typedef struct MEMCHUNK_S
{
	unsigned long       size;
	MEMPTR_T            link;
	unsigned long       type;
	unsigned long       csum;
} MEMCHUNK_T;

#define	CHUNK_MAGIC         0xB562439Fu
#define	CHUNK_SHIFT         4

#define	ALIGN_ADD           0x0000000FL
#define	ALIGN_AND          ~0x0000000FL

#define	TYPE_USED           0x80000000
#define	TIME_ADD            0x00010000

//
// Static Variables.
//

MEMCHUNK_T *        pcl__gFreeList = NULL;

MEMCHUNK_T *        pcl__gHeapBtm = NULL;
MEMCHUNK_T *        pcl__gHeapTop = NULL;

unsigned long       ul___gHeapTimeStamp;

long                sl___gCurMemFree;
long                sl___gCurMemUsed;
long                sl___gMaxMemUsed;

//----------------------------------------------------------------------
// Interrupt Control Macros.
//
// Assembly versions of SwEnterCriticalSection() and SwExitCriticalSection().
// Unlike the library routines, these preserve the original interrupt
// setting, and so can safely be called in a nested fashion.
//----------------------------------------------------------------------

#define di(localvar) {__asm__ volatile ("mfc0 %0,$12; addiu $2,$0,-0x402; and $2,$2,%0; mtc0 $2,$12; nop" : "=r" (localvar) : "0" (localvar) : "$2", "cc"); }
#define ei(localvar) {__asm__ volatile ("mtc0 %0,$12; nop" :: "r" (localvar) : "cc"); }

/*-------------------------------------------------------------
  ShowMemLeft ~ Chris Deel
-------------------------------------------------------------*/
unsigned long ShowMemLeft(void)
{
	#if PC_DEBUG
	printf("Used:%dk Max:%dk   Free: %dk\n",
			sl___gCurMemUsed >> (10 - CHUNK_SHIFT),
			sl___gMaxMemUsed >> (10 - CHUNK_SHIFT),
			sl___gCurMemFree >> (10 - CHUNK_SHIFT) );
	#endif PC_DEBUG

	return sl___gCurMemFree;
}


// **************************************************************************
// * TestHeap ()                                                            *
// **************************************************************************
// * Test the heap for corruption                                           *
// **************************************************************************
// * Inputs  -                                                              *
// *                                                                        *
// * Output  SI    0 if OK, NZ if an error is discovered                    *
// **************************************************************************

int TestHeap (void)
{
	// Local variables.
	MEMCHUNK_T *        pcl__Btm;
	MEMCHUNK_T *        pcl__Top;
	MEMCHUNK_T *        pcl__Cur;
	MEMCHUNK_T *        pcl__Lst;

	unsigned long *     pul__i;
	int                 si___i;

	int                 SR;

	// Disable interrupts before continuing since this code is not re-entrant.

	di(SR);

	// Check the integrity of the memory blocks.

	pcl__Cur =
	pcl__Btm = pcl__gHeapBtm;
	pcl__Top = pcl__gHeapTop;

	while ((pcl__Cur >= pcl__Btm) && (pcl__Cur < pcl__Top))
	{
		if (((unsigned long *) pcl__Cur)[0] +
			((unsigned long *) pcl__Cur)[1] +
			((unsigned long *) pcl__Cur)[2] +
			((unsigned long *) pcl__Cur)[3] != CHUNK_MAGIC)
		{
			ei(SR);
			return (1);
		}
		pcl__Cur += pcl__Cur->size;
	}

	if (((unsigned long *) pcl__Cur)[0] +
		((unsigned long *) pcl__Cur)[1] +
		((unsigned long *) pcl__Cur)[2] +
		((unsigned long *) pcl__Cur)[3] != CHUNK_MAGIC)
	{
		ei(SR);
		return (1);
	}

	// Check the integrity of the free list links.
	//
	// Force a timeout of 1024 free blocks (massive fragmentation).

	pcl__Lst = pcl__gFreeList;
	pcl__Cur = pcl__Lst->link.next;

	for (si___i = 1024; si___i != 0; si___i -= 1)
	{
		if ((pcl__Cur <  pcl__Btm) ||
			(pcl__Cur >= pcl__Top) ||
			((((unsigned int) pcl__Cur) & 15) != 0) ||
			(pcl__Cur == pcl__Lst)) break;
		pcl__Cur = pcl__Cur->link.next;
	}

	if (pcl__Cur != pcl__Lst)
	{
		ei(SR);
		return (1);
	}

	// Restore interrupt status and return result.

	ei(SR);

	return (0);
}



// **************************************************************************
// * DumpHeap ()                                                            *
// **************************************************************************
// * Print out the memory heap list to the debugger message window          *
// **************************************************************************
// * Inputs  -                                                              *
// *                                                                        *
// * Output  -                                                              *
// **************************************************************************

void DumpHeap (void)
{
  #if PC_DEBUG
	// Local variables.
	MEMCHUNK_T *        pcl__Loop;
	MEMCHUNK_T *        pcl__List;
	long                sl___Free = 0;
	int                 si___i    = 0;

	// Go through the heap printing up the blocks.

	printf("\n");

	pcl__List = pcl__gHeapBtm;

	while (pcl__List < pcl__gHeapTop)
	{
		if (((unsigned long *) pcl__List)[0] +
			((unsigned long *) pcl__List)[1] +
			((unsigned long *) pcl__List)[2] +
			((unsigned long *) pcl__List)[3] != CHUNK_MAGIC)
		{
			printf("Addr 0x%08X Size ???????? - Corrupt\n",
				pcl__List);
			break;
		}
		else if (((signed long) pcl__List->type) >= 0)
		{
			printf("Addr 0x%08X Size 0x%06X - Free\n",
				pcl__List,
				pcl__List->size << CHUNK_SHIFT);
			si___i    += 1;
			sl___Free += pcl__List->size;
		}
		else
		{
			char * pcz__File;
			char * pcz__Temp;
			printf("Addr 0x%08X Size 0x%06X - ",
				pcl__List,
				pcl__List->size << CHUNK_SHIFT);
			if ((pcz__File = pcl__List->link.file) == NULL)
			{
				printf("Time %d\n",
					(pcl__List->type & 0x7FFF0000) >> 16);
			}
			else
			{
				pcz__Temp = pcz__File;
				while (*pcz__Temp++ != 0);
				while (--pcz__Temp >= pcz__File)
				{
					if (*pcz__Temp == '\\') break;
					if (*pcz__Temp == '/')  break;
				}
				pcz__Temp += 1;
				printf("Time %d File %s #%d\n",
					(pcl__List->type & 0x7FFF0000) >> 16,
					pcz__Temp,
					(pcl__List->type & 0x0000FFFF) );
			}
		}
		pcl__List += pcl__List->size;
	}

	// Print out the total memory usage.

	if (pcl__List == pcl__gHeapTop)
	{
		printf("\nTotal Memory Used = 0x%08X\n", sl___gCurMemUsed << CHUNK_SHIFT);
		printf("Total Memory Free = 0x%08X (%dk) in %d block(s)\n",
			sl___Free << CHUNK_SHIFT,
			sl___Free >> (10 - CHUNK_SHIFT),
			si___i);
	}

	// All done.

	printf("\n");
	return;

  #endif // PC_DEBUG
}


// **************************************************************************
// * InitHeap ()                                                            *
// **************************************************************************
// * Initialize the memory heap                                             *
// **************************************************************************
// * Inputs  UL *  Ptr to start of free memory                              *
// *         UL    Size of free memory (which I ignore)                     *
// *                                                                        *
// * Output  -                                                              *
// *                                                                        *
// * N.B.    The memory definitions given at the start of this source file  *
// *         override the memory size that is passed.                       *
// *                                                                        *
// *         One of the reasons for doing this is that I can't seem to      *
// *         change PsyQ's default 32KB stack size which is way too large.  *
// **************************************************************************

extern void *MyHeapBase;

void InitHeap (unsigned long * pul__MemAddr, unsigned long ul___MemSize)
{

	// Local variables.
	MEMCHUNK_T *        pcl__List;

	unsigned long       ul___Size;
	MEMPTR_T            cl___Link;
	unsigned long       ul___Type;
	unsigned long       ul___Csum;

	// Correct for PsyQ startup code in __SN_ENTRY_POINT writing 8 bytes
	// beyond the end of the stack.
	//
	//	asm("lui   $sp,0x8020");
	//	asm("addiu $sp,$sp,-32");
	//
	// Since this has now been fixed, I've removed the stack setup from
	// here, and let the PsyQ startup code do it again.

	// Fatal error !
	//
	// I could probably find some #pragma way of testing for this, but
	// I'm too busy right now.


	if (sizeof(MEMCHUNK_T) != 16)
	{
		#if	PC_DEBUG
			printf("Fatal Error - MEMCHUNK_T is not 16 bytes !\n");
		#endif
		while (1);
	}

	// Allocate and track memory usage in 16 byte chunks.

	cl___Link.data = ((((unsigned long) pul__MemAddr) + ALIGN_ADD) & ALIGN_AND);

	#if MEM_END
	ul___Size  = ((MEM_END - cl___Link.data) >> CHUNK_SHIFT) - 1;
	#else
	ul___Size  = (ul___MemSize >> CHUNK_SHIFT) - 2;
	#endif

	ul___Type      = 0;
	ul___Csum      = CHUNK_MAGIC - (ul___Size + cl___Link.data + ul___Type);

	pcl__List        =
	pcl__gFreeList   = cl___Link.next;
	sl___gCurMemFree = ul___Size;
	sl___gCurMemUsed =
	sl___gMaxMemUsed = 0;

	pcl__List->size      = ul___Size;
	pcl__List->link.next = cl___Link.next;
	pcl__List->type      = 0;
	pcl__List->csum      = CHUNK_MAGIC - (cl___Link.data + ul___Size);

	pcl__gHeapBtm = pcl__List;

	pcl__List += ul___Size;

	pcl__List->size      = 0;
	pcl__List->link.next = cl___Link.next;
	pcl__List->type      = 0;
	pcl__List->csum      = CHUNK_MAGIC - (cl___Link.data);

	pcl__gHeapTop = pcl__List;

	ul___gHeapTimeStamp = TIME_ADD;

	// All done.
  	#if PC_DEBUG
 	 printf("Init Mem: %X %X %X  ",
 	 		pul__MemAddr, ul___MemSize,
 	 			(unsigned long)pul__MemAddr + ul___MemSize );
 	 ShowMemLeft();
  	#endif
	return;
}



// **************************************************************************
// * malloc ()                                                              *
// **************************************************************************
// * Allocate a block of memory                                             *
// **************************************************************************
// * Inputs  size_t  # of bytes to allocate                                 *
// *                                                                        *
// * Output  void *  Ptr to memory allocated or NULL if none left           *
// **************************************************************************

void * malloc (size_t ul___Size)
{
	return (mallocfl(ul___Size, NULL, 0));
}

void * mallocfl (size_t ul___Need, char * pcz__File, unsigned long ul___Line)
{

	// Local variables.
	MEMCHUNK_T *        pcl__Loop;
	MEMCHUNK_T *        pcl__Last;
	MEMCHUNK_T *        pcl__This;

	unsigned long       ul___Size;

	int                 SR;

	//

	if (ul___Need == 0)
	{
		return (NULL);
	}

	// Disable interrupts before continuing since this code is not re-entrant.

	di(SR);

	ul___Need =
		((ul___Need+sizeof(MEMCHUNK_T)-1)>>CHUNK_SHIFT)+1;

	pcl__Loop =
	pcl__Last = pcl__gFreeList;
	pcl__This = pcl__Last->link.next;

	// Search for a block big enough.

	while (1)
	{
		ul___Size = pcl__This->size;

		if (ul___Size < ul___Need)
		{
			// Try next block.
			if (pcl__This == pcl__Loop) break;
			pcl__Last = pcl__This;
			pcl__This = pcl__This->link.next;
			continue;
		}
		else
		{
			if (ul___Size == ul___Need)
			{
				// Allocate this entire block (unless only free block).
				if (pcl__This == pcl__Last) break;
				ul___Size            = (unsigned long)
				pcl__Last->link.next = pcl__This->link.next;
				pcl__Last->csum      = CHUNK_MAGIC -
					(pcl__Last->size + ul___Size + pcl__Last->type);
			}
			else
			{
				// Allocate part of this block.
				ul___Size      -= ul___Need;
				pcl__This->size = ul___Size;
				pcl__This->csum = CHUNK_MAGIC -
					(ul___Size + pcl__This->link.data + pcl__This->type);
				pcl__This      += ul___Size;
			}

			ul___Line |= (TYPE_USED | ul___gHeapTimeStamp);

			pcl__This->size      = ul___Need;
			pcl__This->link.file = pcz__File;
			pcl__This->type      = ul___Line;
			pcl__This->csum = CHUNK_MAGIC -
				(ul___Need + ((unsigned long) pcz__File) + ul___Line);

			pcl__gFreeList       = pcl__Last;
			sl___gCurMemFree    -= ul___Need;
			sl___gCurMemUsed    += ul___Need;

			if (sl___gMaxMemUsed < sl___gCurMemUsed)
			{
				sl___gMaxMemUsed = sl___gCurMemUsed;
			}

			ul___gHeapTimeStamp += TIME_ADD;

			// Restore interrupt status and return result.

			ei(SR);

			return ((void *) (pcl__This + 1));
		}
	}

	// Return failure.

	ei(SR);

	#if DEBUG
	printf("Malloc Error");
	printf("\nCan't allocate 0x%08X bytes !\n", ul___Need << CHUNK_SHIFT);
	DumpHeap();
	 asm("break 0");
	 getch();

	#endif

	return (NULL);
}



// **************************************************************************
// * calloc ()                                                              *
// **************************************************************************
// * Allocate and clear a block of memory                                   *
// **************************************************************************
// * Inputs  size_t  # of elements to allocate                              *
// *         size_t  Size of each element in bytes                          *
// *                                                                        *
// * Output  void *  Ptr to memory allocated or NULL if none left           *
// **************************************************************************

extern void MemClrLinePara (void *, unsigned int);

void * calloc (size_t ui___NumElem, size_t ui___LenElem)
{
	return (callocfl(ui___NumElem, ui___LenElem, NULL, 0));
}

void * callocfl (size_t ul___NumElem, size_t ul___LenElem, char * pcz__File, unsigned long ul___Line)
{
  #if CALLOC_USED
	// Local variables.
	unsigned long       ul___Size;

	MEMCHUNK_T *        pcl__This;

	//

	if (ul___LenElem == 1)
	{
		ul___Size = ul___NumElem;
	} else
	{
		ul___Size = ul___NumElem * ul___LenElem;
	}

	if (ul___Size == 0)
	{
		return (NULL);
	}

	pcl__This = (MEMCHUNK_T *) mallocfl(ul___Size, pcz__File, ul___Line);

	if (pcl__This != NULL)
	{
		#if 1
		unsigned int * pui__Tmp;
		unsigned int   ui___Tmp;
		ui___Tmp = (pcl__This-1)->size - 1;
		pui__Tmp = (unsigned int *) pcl__This;
		do
		{
			pui__Tmp[0] = 0;
			pui__Tmp[1] = 0;
			pui__Tmp[2] = 0;
			pui__Tmp[3] = 0;
			pui__Tmp   += 4;
		} while (--ui___Tmp);
		#else
		MemClrLinePara(pcl__This, (pcl__This-1)->size - 1);
		#endif
	}

	return (pcl__This);
  #else
  	printf("NO calloc\n");
  	return NULL;
  #endif // CALLOC_USED
}



// **************************************************************************
// * free ()                                                                *
// **************************************************************************
// * Free a block of memory                                                 *
// **************************************************************************
// * Inputs  void *  Ptr to memory block to free                            *
// *                                                                        *
// * Output  -                                                              *
// **************************************************************************

void free (void * pbf__Chunk)
{

	// Local variables.
	MEMCHUNK_T *        pcl__Chunk;
	MEMCHUNK_T *        pcl__This;
	MEMCHUNK_T *        pcl__Next;

	char *              pcz__Error;

	int                 SR;

	// Disable interrupts before continuing since this code is not re-entrant.

	di(SR);

	// Make sure that the block is valid.

	pcl__Chunk = ((MEMCHUNK_T *) pbf__Chunk) - 1;
	{
		unsigned long * pul__i;

		pul__i = (unsigned long *) pcl__Chunk;

		if ((pul__i[0] + pul__i[1] + pul__i[2] + pul__i[3]) != CHUNK_MAGIC)
		{
			goto errorInvalid;
		}

		pul__i = (unsigned long *) (pcl__Chunk+pcl__Chunk->size);

		if ((pul__i[0] + pul__i[1] + pul__i[2] + pul__i[3]) != CHUNK_MAGIC)
		{
			goto errorOverrun;
		}

		if (((signed long) pcl__Chunk->type) >= 0)
		{
			goto errorFree;
		}
	}

	// Mark this header as unused to ensure that freeing it twice will trap.

	pcl__Chunk->type = 0;

	// Now free up the block.

	sl___gCurMemFree += pcl__Chunk->size;
	sl___gCurMemUsed -= pcl__Chunk->size;

	pcl__This = pcl__gFreeList;
	pcl__Next = pcl__This->link.next;

	// Search the free block list for the blocks on either side of the
	// block that we're freeing.

	if (pcl__This != pcl__Next)
	{
		while (1)
		{
			if ((pcl__Chunk > pcl__This)  &&
				(pcl__Chunk < pcl__Next)) break;

			if (pcl__This >= pcl__Next)
			{
				if ((pcl__Chunk > pcl__This)  ||
					(pcl__Chunk < pcl__Next)) break;
			}
			pcl__This = pcl__Next;
			pcl__Next = pcl__This->link.next;
		}
	}

	// Merge contiguous free space.

	if ((pcl__Chunk + pcl__Chunk->size) == pcl__Next)
	{
		pcl__Chunk->size     += pcl__Next->size;
		pcl__Chunk->link.next = pcl__Next->link.next;
		if (pcl__This == pcl__Next)
		{
			pcl__This  = pcl__Chunk;
		}
	} else
	{
		pcl__Chunk->link.next = pcl__Next;
	}

	if ((pcl__This + pcl__This->size) == pcl__Chunk)
	{
		pcl__This->size     += pcl__Chunk->size;
		pcl__This->link.next = pcl__Chunk->link.next;
		pcl__Chunk = pcl__This;
	} else
	{
		pcl__This->link.next = pcl__Chunk;
		pcl__This->csum      = CHUNK_MAGIC -
			(pcl__This->size + ((unsigned long) pcl__Chunk) + pcl__This->type);
	}

	pcl__Chunk->csum = CHUNK_MAGIC -
		(pcl__Chunk->size + pcl__Chunk->link.data + pcl__Chunk->type);

	pcl__gFreeList = pcl__This;

	// Restore interrupt status and return.

	ei(SR);

	return;

	// Error conditions (reached via the dreaded GOTO).

errorInvalid:
	pcz__Error = "WARNING - Attempt to free invalid or corrupt memory block!";
	goto errorExit;
errorOverrun:
	pcz__Error = "WARNING - Memory overrun detected during free()!";
	goto errorExit;
errorFree:
	pcz__Error = "WARNING - Attempt to free an already free memory block!";
errorExit:
		// I've done it this way to allow you to put a single breakpoint
		// here to catch all of the error conditions.
	ei(SR);
	#if PC_DEBUG
	 printf("%s  %08X\n", pcz__Error, pbf__Chunk);
	 DumpHeap();
	 asm("break 0");
//	 getch();
	#endif
	return;
}



// **************************************************************************
// * realloc ()                                                             *
// **************************************************************************
// * Reallocate a memory block                                              *
// **************************************************************************
// * Inputs  void *  Ptr to old memory block                                *
// *         size_t  Size of new memory block                               *
// *                                                                        *
// * Output  void *  Ptr to new memory block                                *
// *                                                                        *
// * N.B.    This is a pretty shitty implementation, but it does meet the   *
// *         ANSI definition.                                               *
// *                                                                        *
// *         It would be more useful if it was rewritten to allow a block   *
// *         to be shrunk in place, but even then you'd still have to code  *
// *         very, very carefully to avoid heap fragmentation.              *
// *                                                                        *
// *         There is also a potential problem here in that if you realloc  *
// *         from a small block to a large block, the code that copies the  *
// *         old data into the new block will overrun the end of the old    *
// *         block, possibly triggering a page fault if it is run in a      *
// *         system with memory management hardware.                        *
// *                                                                        *
// *         My apologies, but since I never use realloc(), I can't be      *
// *         bothered to provide a better implementation.                   *
// **************************************************************************

void * realloc (void * pbf__Old, size_t ui___Size)
{
	return (reallocfl(pbf__Old, ui___Size, NULL, 0));
}

void * reallocfl (void * pbf__Old, size_t ui___Size, char * pcz__File, unsigned long ul___Line)
{
  #if REALLOC_USED
	// Local variables.
	int *               psi__New = NULL;
	int *               psi__Tmp;

	//

	if (ui___Size == 0)
	{
		if (pbf__Old != NULL) free(pbf__Old);
	}
	else
	{
		if ((psi__New = (int *) mallocfl(ui___Size, pcz__File, ul___Line)) != NULL)
		{
			if (pbf__Old != NULL)
			{
				ui___Size = (ui___Size+sizeof(MEMCHUNK_T)- 1)>>CHUNK_SHIFT;
				ui___Size = ui___Size * (sizeof(MEMCHUNK_T)/sizeof(int));

				psi__Tmp  = psi__New;
				while (ui___Size--)
				{
					*psi__Tmp++ = *((int *) pbf__Old)++;
				}

				free(pbf__Old);
			}
		}
	}

	return ((void *) psi__New);

  #else // realloc not used
  	printf("NO realloc!!\n");
	return NULL;
  #endif
}

void PrintHeap(void)
{
	DebugPrintf("Addr 0x%08X",pcl__gHeapBtm);
	DebugPrintf("Addr 0x%08X",pcl__gHeapTop);
}

// **************************************************************************

